/// The code in this file is derived from the Dart SDK [here](https://github.com/dart-lang/sdk/blob/master/sdk/lib/collection/maps.dart) written by the Dart project authors.

import 'package:meta/meta.dart';

typedef V IfAbsentHandler<V>();

/// Wrapper around a class that implements [Map] that only exposes `Map`
/// members.
///
/// A simple wrapper that delegates all `Map` members to the map provided in the
/// constructor.
///
/// Base for delegating map implementations like [UnmodifiableMapView].
class MapView<K, V> implements Map<K, V> {
  @protected
  final Map<K, V> internalMap;

  /// Creates a new instance of `MapView`
  const MapView(Map<K, V> map) : internalMap = map;

  Map<RK, RV> cast<RK, RV>() => internalMap.cast<RK, RV>();
  V operator [](Object key) => internalMap[key];
  void operator []=(K key, V value) {
    internalMap[key] = value;
  }

  void addAll(Map<K, V> other) {
    internalMap.addAll(other);
  }

  void clear() {
    internalMap.clear();
  }

  V putIfAbsent(K key, V ifAbsent()) => internalMap.putIfAbsent(key, ifAbsent);
  bool containsKey(Object key) => internalMap.containsKey(key);
  bool containsValue(Object value) => internalMap.containsValue(value);
  void forEach(void action(K key, V value)) {
    internalMap.forEach(action);
  }

  bool get isEmpty => internalMap.isEmpty;
  bool get isNotEmpty => internalMap.isNotEmpty;
  int get length => internalMap.length;
  Iterable<K> get keys => internalMap.keys;
  V remove(Object key) => internalMap.remove(key);
  String toString() => internalMap.toString();
  Iterable<V> get values => internalMap.values;

  Iterable<MapEntry<K, V>> get entries => internalMap.entries;

  void addEntries(Iterable<MapEntry<K, V>> entries) {
    internalMap.addEntries(entries);
  }

  Map<K2, V2> map<K2, V2>(MapEntry<K2, V2> transform(K key, V value)) => internalMap.map<K2, V2>(transform);

  V update(K key, V update(V value), {IfAbsentHandler<V> ifAbsent}) => internalMap.update(key, update, ifAbsent: ifAbsent);

  void updateAll(V update(K key, V value)) {
    internalMap.updateAll(update);
  }

  void removeWhere(bool test(K key, V value)) {
    internalMap.removeWhere(test);
  }
}

/// Mixin that overrides mutating map operations with implementations that
/// throw.
///
/// Copy of Dart's internal `_UnmodifiableMapMixin`
mixin UnmodifiableMapMixin<K, V> on Map<K, V> {
  /// This operation is not supported by an unmodifiable map.
  void operator []=(K key, V value) {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }

  /// This operation is not supported by an unmodifiable map.
  void addAll(Map<K, V> other) {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }

  /// This operation is not supported by an unmodifiable map.
  void addEntries(Iterable<MapEntry<K, V>> entries) {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }

  /// This operation is not supported by an unmodifiable map.
  void clear() {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }

  /// This operation is not supported by an unmodifiable map.
  V remove(Object key) {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }

  /// This operation is not supported by an unmodifiable map.
  void removeWhere(bool test(K key, V value)) {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }

  /// This operation is not supported by an unmodifiable map.
  V putIfAbsent(K key, V ifAbsent()) {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }

  /// This operation is not supported by an unmodifiable map.
  V update(K key, V update(V value), {IfAbsentHandler<V> ifAbsent}) {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }

  /// This operation is not supported by an unmodifiable map.
  void updateAll(V update(K key, V value)) {
    throw UnsupportedError('Cannot modify unmodifiable map');
  }
}

/// View of a [Map] that disallow modifying the map.
///
/// A wrapper around a `Map` that forwards all members to the map provided in
/// the constructor, except for operations that modify the map.
/// Modifying operations throw instead.
class UnmodifiableMapView<K, V> extends MapView<K, V> with UnmodifiableMapMixin<K, V> {
  const UnmodifiableMapView(Map<K, V> map) : super(map);

  Map<RK, RV> cast<RK, RV>() => UnmodifiableMapView<RK, RV>(internalMap.cast<RK, RV>());
}
