import 'package:plexus/expose.dart';
import 'package:plexus/plexus.dart';

class Localizations extends UnmodifiableMapView<Locale, Localization> {
  const Localizations(this.fallbackLocale, this.supportedLanguages, this.supportedLocales, Map<Locale, Localization> localizations) : super(localizations);

  final Set<Locale> supportedLocales;
  final Set<Language> supportedLanguages;
  final Locale fallbackLocale;

  bool isLocaleSupported(Locale locale) => supportedLocales.contains(locale);
  bool isLanguageSupported(Language language) => supportedLanguages.contains(language);

  /// Prefer this to the indexing operator, which should only be used for locales that are guaranteed to exist.
  Localization operator |(LanguageBase language) => language is Locale ? (isLocaleSupported(language) ? internalMap[language] : isLanguageSupported(language.language)) : isLanguageSupported(language) ? internalMap[language] : internalMap[fallbackLocale];
}

class Localization extends UnmodifiableMapView<String, dynamic> {
  const Localization(this.locale, Map<String, dynamic> localizations) : super(localizations);

  final Locale locale;

  /// Prefer this to the indexing operator, which should only be used for keys that are guaranteed to exist.
  dynamic operator |(String key) => internalMap.containsKey(key) ? internalMap[key] : key;

  static const ENGLISH_COMMON = Localization(Locale(Locale.LANGUAGE_ENGLISH), {
    'greeting': 'Welcome',
    'enthusiastic_greeting': 'Welcome!'
  });

  static const ENGLISH_CANADA = Localization(Locale(Locale.LANGUAGE_ENGLISH, Locale.COUNTRY_CANADA), ENGLISH__CANADA);

  static const ENGLISH__COMMON = {
    'greeting': 'Welcome',
    'enthusiastic_greeting': 'Welcome!'
  };

  static const ENGLISH__CANADA = {
    ...ENGLISH__COMMON,
  };
}

class Locale extends LanguageBase {
  const Locale(this.language, [this.country]);

  final Language language;

  @nullable
  final Country country;

  @override
  String get code => language.code;

  @override
  String get label => language.label;

  static const COUNTRY_CANADA = Country('CA', 'Canada');
  static const COUNTRY_UNITED_KINGDOM = Country('UK', 'United Kingdom');
  static const COUNTRY_UNITED_STATES = Country('US', 'United States of America');

  static const LANGUAGE_ENGLISH = Language('en', 'English');

  static const languages = {
    LANGUAGE_ENGLISH
  };

  static const countries = {
    COUNTRY_CANADA,
    COUNTRY_UNITED_KINGDOM,
    COUNTRY_UNITED_STATES,
  };

  static const locales = {
    Locale(LANGUAGE_ENGLISH, COUNTRY_CANADA),
    Locale(LANGUAGE_ENGLISH, COUNTRY_UNITED_KINGDOM),
    Locale(LANGUAGE_ENGLISH, COUNTRY_UNITED_STATES)
  };
}

abstract class LanguageBase {
  const LanguageBase();

  String get code;
  String get label;
}

class Language extends LanguageBase {
  const Language(this.code, this.label);

  @override
  final String code;

  @override
  final String label;
}

class Country {
  const Country(this.code, this.label);

  final code;
  final label;
}
